#
# MIT License
#
# Copyright (c) 2022 KuFlow
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

from datetime import date, datetime
from typing import Any, Dict, List, Optional

from .._generated.models import ProcessElementValue
from ..models import JsonFormsFile, JsonFormsPrincipal, JsonFormsValue, Process
from .element_values import (
    ElementValueSimpleType,
    ElementValueUnion,
    ProcessElementValueAccessor,
    add_element_value,
    add_element_value_list,
    find_element_value_as_date,
    find_element_value_as_float,
    find_element_value_as_str,
    get_element_value_as_date,
    get_element_value_as_date_list,
    get_element_value_as_float,
    get_element_value_as_float_list,
    get_element_value_as_str,
    get_element_value_as_str_list,
    get_element_value_valid,
    get_element_value_valid_at,
    set_element_value,
    set_element_value_list,
    set_element_value_valid,
    set_element_value_valid_at,
)
from .json_forms import (
    JsonFormDataAccessor,
    find_json_forms_property_as_date,
    find_json_forms_property_as_datetime,
    find_json_forms_property_as_dict,
    find_json_forms_property_as_float,
    find_json_forms_property_as_int,
    find_json_forms_property_as_json_forms_file,
    find_json_forms_property_as_json_forms_principal,
    find_json_forms_property_as_list,
    find_json_forms_property_as_str,
    get_json_forms_property_as_date,
    get_json_forms_property_as_datetime,
    get_json_forms_property_as_dict,
    get_json_forms_property_as_float,
    get_json_forms_property_as_int,
    get_json_forms_property_as_json_forms_file,
    get_json_forms_property_as_json_forms_principal,
    get_json_forms_property_as_list,
    get_json_forms_property_as_str,
)


class ProcessUtils:
    @staticmethod
    def get_element_value_valid(process: Process, element_definition_code: str) -> bool:
        """
        Check if all related valid values are TRUE.

        Arguments:
            process: The process.
            element_definition_code: Element Definition Code.

        Returns:
            True if all related valid values are TRUE, otherwise False.
        """
        return get_element_value_valid(CurrentElementValueAccessor(process, element_definition_code))

    @staticmethod
    def get_element_value_valid_at(
        process: Process,
        element_definition_code: str,
        index: int,
    ) -> bool:
        """
        Check if the requested valid value at the given index is TRUE.

        Arguments:
            process: The process.
            element_definition_code: The element definition code.
            index: The element value index.

        Returns:
            The requested valid value if it exists, otherwise None.
        """
        return get_element_value_valid_at(CurrentElementValueAccessor(process, element_definition_code), index)

    @staticmethod
    def set_element_value_valid(process: Process, element_definition_code: str, valid: Optional[bool]) -> Process:
        """
        Set the valid value for all element values.

        Arguments:
            process: The process.
            element_definition_code: The element definition code.
            valid: The valid value.

        Returns:
            The passed process.
        """
        set_element_value_valid(CurrentElementValueAccessor(process, element_definition_code), valid)

        return process

    @staticmethod
    def set_element_value_valid_at(
        process: Process,
        element_definition_code: str,
        valid: Optional[bool],
        index: int,
    ) -> Process:
        """
        Set the valid value for the selected element value.

        Arguments:
            process: The process.
            element_definition_code: The element definition code.
            valid: The valid value.
            index: The element value index.

        Returns:
            The passed process.
        """
        set_element_value_valid_at(CurrentElementValueAccessor(process, element_definition_code), valid, index)

        return process

    @staticmethod
    def set_element_value(
        process: Process,
        element_definition_code: str,
        element_value: Optional[ElementValueSimpleType] = None,
    ) -> Process:
        """
        Set an element value.

        Arguments:
            process: The process.
            element_definition_code: The element definition code.
            element_value: The element value. If the value is None, all current values are removed.

        Returns:
            The passed process.
        """
        set_element_value(CurrentElementValueAccessor(process, element_definition_code), element_value)

        return process

    @staticmethod
    def set_element_value_list(
        process: Process,
        element_definition_code: str,
        element_values: Optional[List[ElementValueSimpleType]] = None,
    ) -> Process:
        """
        Set an element value.

        Arguments:
            process: The process.
            element_definition_code: The element definition code.
            element_values: The element value. If the value is None or empty, all current values are removed.

        Returns:
            The passed process.
        """
        set_element_value_list(
            CurrentElementValueAccessor(process, element_definition_code),
            element_values,
        )

        return process

    @staticmethod
    def add_element_value(
        process: Process,
        element_definition_code: str,
        element_value: Optional[ElementValueSimpleType] = None,
    ) -> Process:
        """
        Add an element value.

        Arguments:
            process: The process.
            element_definition_code: The element definition code.
            element_value: The element value. If the value is None, all current values are removed.

        Returns:
            The passed process.
        """
        add_element_value(CurrentElementValueAccessor(process, element_definition_code), element_value)

        return process

    @staticmethod
    def add_element_value_list(
        process: Process,
        element_definition_code: str,
        element_values: Optional[List[ElementValueSimpleType]] = None,
    ) -> Process:
        """
        Add element values.

        Arguments:
            process: The process.
            element_definition_code: The element definition code.
            element_values: The element values. If the values is None or empty, all current values are removed.

        Returns:
            The passed model related object.
        """
        add_element_value_list(
            CurrentElementValueAccessor(process, element_definition_code),
            element_values,
        )

        return process

    @staticmethod
    def get_element_value_as_str(process: Process, element_definition_code: str) -> str:
        """
        Get an element as a str.

        Arguments:
            process: The process.
            element_definition_code: The element definition code.

        Returns:
            The element value as a str.
        """
        return get_element_value_as_str(CurrentElementValueAccessor(process, element_definition_code))

    @staticmethod
    def find_element_value_as_str(process: Process, element_definition_code: str) -> Optional[str]:
        """
        Try to get an element as a str.

        Arguments:
            process: The process.
            element_definition_code: The element definition code.

        Returns:
            The element value as a str.
        """
        return find_element_value_as_str(CurrentElementValueAccessor(process, element_definition_code))

    @staticmethod
    def get_element_value_as_str_list(process: Process, element_definition_code: str) -> List[str]:
        """
        Try to get an element as a str list.

        Arguments:
            process: The process.
            element_definition_code: The element definition code.

        Returns:
            The element values as a str list.
        """
        return get_element_value_as_str_list(CurrentElementValueAccessor(process, element_definition_code))

    @staticmethod
    def get_element_value_as_float(process: Process, element_definition_code: str) -> float:
        """
        Get an element as a float.

        Arguments:
            process: The process.
            element_definition_code: The element definition code.

        Returns:
            The element value as a float.
        """
        return get_element_value_as_float(CurrentElementValueAccessor(process, element_definition_code))

    @staticmethod
    def find_element_value_as_float(process: Process, element_definition_code: str) -> float:
        """
        Try to get an element as a float.

        Arguments:
            process: The process.
            element_definition_code: The element definition code.

        Returns:
            The element value as a float.
        """
        return find_element_value_as_float(CurrentElementValueAccessor(process, element_definition_code))

    @staticmethod
    def get_element_value_as_float_list(process: Process, element_definition_code: str) -> List[float]:
        """
        Get all elements as a float list.

        Arguments:
            process: The process.
            element_definition_code: The element definition code.

        Returns:
            The element value as a float.
        """
        return get_element_value_as_float_list(CurrentElementValueAccessor(process, element_definition_code))

    @staticmethod
    def get_element_value_as_date(process: Process, element_definition_code: str) -> date:
        """
        Get an element as a date.

        Arguments:
            process: The process.
            element_definition_code: The element definition code.

        Returns:
            The element value as a date.
        """
        return get_element_value_as_date(CurrentElementValueAccessor(process, element_definition_code))

    @staticmethod
    def find_element_value_as_date(process: Process, element_definition_code: str) -> Optional[date]:
        """
        Try to get  an element as a date.

        Arguments:
            process: The process.
            element_definition_code: The element definition code.

        Returns:
            The element value as a date.
        """
        return find_element_value_as_date(CurrentElementValueAccessor(process, element_definition_code))

    @staticmethod
    def get_element_value_as_date_list(process: Process, element_definition_code: str) -> List[date]:
        """
        Get all elements as date list.

        Arguments:
            process: The process.
            element_definition_code: The element definition code.

        Returns:
            The element value as a date.
        """
        return get_element_value_as_date_list(CurrentElementValueAccessor(process, element_definition_code))

    @staticmethod
    def get_entity_property_as_str(process: Process, property_path: str) -> str:
        """
        Get a json property as "str" following the "propertyPath" passed.

        Arguments:
            process: The process.
            property_path: Property path to find. ie: "user.name" or "users.0.name"

        Return:
            the property value if exists.

        Raises:
            ValueError: If property value doesn't exist or has incorrect format
        """
        return get_json_forms_property_as_str(CurrentJsonFormDataAccessor(process), property_path)

    @staticmethod
    def find_entity_property_as_str(process: Process, property_path: str) -> Optional[str]:
        """
        Try to find a json property as "str" following the "propertyPath" passed.

        Arguments:
            process: The process.
            property_path: Property path to find. ie: "user.name" or "users.0.name"

        Return:
            The property value if exists.

        Raises:
            ValueError: If property value has incorrect format
        """
        return find_json_forms_property_as_str(CurrentJsonFormDataAccessor(process), property_path)

    @staticmethod
    def get_entity_property_as_int(process: Process, property_path: str) -> int:
        """
        Get a json property as "int" following the "propertyPath" passed.

        Arguments:
            process: The process.
            property_path: Property path to find. ie: "user.name" or "users.0.name"

        Return:
            the property value if exists.

        Raises:
            ValueError: If property value doesn't exist or has incorrect format
        """
        return get_json_forms_property_as_int(CurrentJsonFormDataAccessor(process), property_path)

    @staticmethod
    def find_entity_property_as_int(process: Process, property_path: str) -> Optional[int]:
        """
        Try to find a json property as "int" following the "propertyPath" passed.

        Arguments:
            process: The process.
            property_path: Property path to find. ie: "user.name" or "users.0.name"

        Return:
            The property value if exists.

        Raises:
            ValueError: If property value has incorrect format
        """
        return find_json_forms_property_as_int(CurrentJsonFormDataAccessor(process), property_path)

    @staticmethod
    def get_entity_property_as_float(process: Process, property_path: str) -> float:
        """
        Get a json property as "float" following the "propertyPath" passed.

        Arguments:
            process: The process.
            property_path: Property path to find. ie: "user.name" or "users.0.name"

        Return:
            the property value if exists.

        Raises:
            ValueError: If property value doesn't exist or has incorrect format
        """
        return get_json_forms_property_as_float(CurrentJsonFormDataAccessor(process), property_path)

    @staticmethod
    def find_entity_property_as_float(process: Process, property_path: str) -> Optional[float]:
        """
        Try to find a json property as "float" following the "propertyPath" passed.

        Arguments:
            process: The process.
            property_path: Property path to find. ie: "user.name" or "users.0.name"

        Return:
            The property value if exists.

        Raises:
            ValueError: If property value has incorrect format
        """
        return find_json_forms_property_as_float(CurrentJsonFormDataAccessor(process), property_path)

    @staticmethod
    def get_entity_property_as_date(process: Process, property_path: str) -> date:
        """
        Get a json property as "date" following the "propertyPath" passed.

        Arguments:
            process: The process.
            property_path: Property path to find. ie: "user.name" or "users.0.name"

        Return:
            the property value if exists.

        Raises:
            ValueError: If property value doesn't exist or has incorrect format
        """
        return get_json_forms_property_as_date(CurrentJsonFormDataAccessor(process), property_path)

    @staticmethod
    def find_entity_property_as_date(process: Process, property_path: str) -> Optional[date]:
        """
        Try to find a json property as "date" following the "propertyPath" passed.

        Arguments:
            process: The process.
            property_path: Property path to find. ie: "user.name" or "users.0.name"

        Return:
            The property value if exists.

        Raises:
            ValueError: If property value has incorrect format
        """
        return find_json_forms_property_as_date(CurrentJsonFormDataAccessor(process), property_path)

    @staticmethod
    def get_entity_property_as_datetime(process: Process, property_path: str) -> datetime:
        """
        Get a json property as "datetime" following the "propertyPath" passed.

        Arguments:
            process: The process.
            property_path: Property path to find. ie: "user.name" or "users.0.name"

        Return:
            the property value if exists.

        Raises:
            ValueError: If property value doesn't exist or has incorrect format
        """
        return get_json_forms_property_as_datetime(CurrentJsonFormDataAccessor(process), property_path)

    @staticmethod
    def find_entity_property_as_datetime(process: Process, property_path: str) -> Optional[datetime]:
        """
        Try to find a json property as "datetime" following the "propertyPath" passed.

        Arguments:
            process: The process.
            property_path: Property path to find. ie: "user.name" or "users.0.name"

        Return:
            The property value if exists.

        Raises:
            ValueError: If property value has incorrect format
        """
        return find_json_forms_property_as_datetime(CurrentJsonFormDataAccessor(process), property_path)

    @staticmethod
    def get_entity_property_as_file(process: Process, property_path: str) -> JsonFormsFile:
        """
        Get a json property as "JsonFormsFile" following the "propertyPath" passed.

        Arguments:
            process: The process.
            property_path: Property path to find. ie: "user.name" or "users.0.name"

        Return:
            the property value if exists.

        Raises:
            ValueError: If property value doesn't exist or has incorrect format
        """
        return get_json_forms_property_as_json_forms_file(CurrentJsonFormDataAccessor(process), property_path)

    @staticmethod
    def find_entity_property_as_file(process: Process, property_path: str) -> Optional[JsonFormsFile]:
        """
        Try to find a json property as "JsonFormsFile" following the "propertyPath" passed.

        Arguments:
            process: The process.
            property_path: Property path to find. ie: "user.name" or "users.0.name"

        Return:
            The property value if exists.

        Raises:
            ValueError: If property value has incorrect format
        """
        return find_json_forms_property_as_json_forms_file(CurrentJsonFormDataAccessor(process), property_path)

    @staticmethod
    def get_entity_property_as_principal(process: Process, property_path: str) -> JsonFormsPrincipal:
        """
        Get a json property as "JsonFormsPrincipal" following the "propertyPath" passed.

        Arguments:
            process: The process.
            property_path: Property path to find. ie: "user.name" or "users.0.name"

        Return:
            the property value if exists.

        Raises:
            ValueError: If property value doesn't exist or has incorrect format
        """
        return get_json_forms_property_as_json_forms_principal(CurrentJsonFormDataAccessor(process), property_path)

    @staticmethod
    def find_entity_property_as_principal(process: Process, property_path: str) -> Optional[JsonFormsPrincipal]:
        """
        Try to find a json property as "JsonFormsPrincipal" following the "propertyPath" passed.

        Arguments:
            process: The process.
            property_path: Property path to find. ie: "user.name" or "users.0.name"

        Return:
            The property value if exists.

        Raises:
            ValueError: If property value has incorrect format
        """
        return find_json_forms_property_as_json_forms_principal(CurrentJsonFormDataAccessor(process), property_path)

    @staticmethod
    def get_entity_property_as_list(process: Process, property_path: str) -> list:
        """
        Get a json property as "list" following the "propertyPath" passed.

        Arguments:
            process: The process.
            property_path: Property path to find. ie: "user.name" or "users.0.name"

        Return:
            the property value if exists.

        Raises:
            ValueError: If property value doesn't exist or has incorrect format
        """
        return get_json_forms_property_as_list(CurrentJsonFormDataAccessor(process), property_path)

    @staticmethod
    def find_entity_property_as_list(process: Process, property_path: str) -> Optional[list]:
        """
        Try to find a json property as "list" following the "propertyPath" passed.

        Arguments:
            process: The process.
            property_path: Property path to find. ie: "user.name" or "users.0.name"

        Return:
            The property value if exists.

        Raises:
            ValueError: If property value has incorrect format
        """
        return find_json_forms_property_as_list(CurrentJsonFormDataAccessor(process), property_path)

    @staticmethod
    def get_entity_property_as_dict(process: Process, property_path: str) -> dict:
        """
        Get a json property as "dict" following the "propertyPath" passed.

        Arguments:
            process: The process.
            property_path: Property path to find. ie: "user.name" or "users.0.name"

        Return:
            the property value if exists.

        Raises:
            ValueError: If property value doesn't exist or has incorrect format
        """
        return get_json_forms_property_as_dict(CurrentJsonFormDataAccessor(process), property_path)

    @staticmethod
    def find_entity_property_as_dict(process: Process, property_path: str) -> Optional[dict]:
        """
        Try to find a json property as "dict" following the "propertyPath" passed.

        Arguments:
            process: The process.
            property_path: Property path to find. ie: "user.name" or "users.0.name"

        Return:
            The property value if exists.

        Raises:
            ValueError: If property value has incorrect format
        """
        return find_json_forms_property_as_dict(CurrentJsonFormDataAccessor(process), property_path)


class CurrentElementValueAccessor(ProcessElementValueAccessor):
    def __init__(self, process: Process, element_definition_code: str):
        self.process = process
        self.element_definition_code = element_definition_code

    def set_element_values(self, element_values: List[ElementValueUnion]):
        if self.process.element_values is None:
            self.process.element_values = {}

        if len(element_values) == 0:
            self.process.element_values.pop(self.element_definition_code, None)
        else:
            self.process.element_values[self.element_definition_code] = element_values

    def get_element_values(self) -> List[ProcessElementValue]:
        if self.process.element_values is None:
            return []

        element_values_by_code = self.process.element_values.get(self.element_definition_code)
        if element_values_by_code is None or len(element_values_by_code) == 0:
            return []

        return element_values_by_code


class CurrentJsonFormDataAccessor(JsonFormDataAccessor):
    def __init__(self, process: Process):
        self.process = process

    def get_data(self) -> Optional[Dict[str, Any]]:
        if self.process.entity is None:
            return None

        return self.process.entity.data

    def set_data(self, data: Optional[Dict[str, Any]]):
        if self.process.entity is None:
            self.process.entity = JsonFormsValue()

        self.process.entity.data = data



# coding=utf-8
# --------------------------------------------------------------------------
#
# MIT License
#
# Copyright (c) 2022 KuFlow
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# --------------------------------------------------------------------------
#
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is regenerated.
#
# --------------------------------------------------------------------------

from ._models import AbstractAudited
from ._models import Authentication
from ._models import AuthenticationEngineCertificate
from ._models import AuthenticationEngineCertificateTls
from ._models import AuthenticationEngineToken
from ._models import DefaultError
from ._models import DefaultErrorInfo
from ._models import JsonFormsValue
from ._models import Log
from ._models import Page
from ._models import PageMetadata
from ._models import Principal
from ._models import PrincipalApplication
from ._models import PrincipalPage
from ._models import PrincipalUser
from ._models import Process
from ._models import ProcessChangeInitiatorCommand
from ._models import ProcessDefinitionSummary
from ._models import ProcessDeleteElementCommand
from ._models import ProcessElementValue
from ._models import ProcessElementValueNumber
from ._models import ProcessElementValueString
from ._models import ProcessPage
from ._models import ProcessPageItem
from ._models import ProcessSaveElementCommand
from ._models import ProcessSaveEntityDataCommand
from ._models import ProcessSaveEntityDocumentResponseCommand
from ._models import RelatedProcess
from ._models import Robot
from ._models import RobotPage
from ._models import RobotSourceFile
from ._models import Task
from ._models import TaskAssignCommand
from ._models import TaskDefinitionSummary
from ._models import TaskDeleteElementCommand
from ._models import TaskDeleteElementValueDocumentCommand
from ._models import TaskElementValue
from ._models import TaskElementValueDocument
from ._models import TaskElementValueDocumentItem
from ._models import TaskElementValueNumber
from ._models import TaskElementValueObject
from ._models import TaskElementValuePrincipal
from ._models import TaskElementValuePrincipalItem
from ._models import TaskElementValueString
from ._models import TaskPage
from ._models import TaskPageItem
from ._models import TaskSaveElementCommand
from ._models import TaskSaveJsonFormsValueDataCommand
from ._models import TaskSaveJsonFormsValueDocumentResponseCommand
from ._models import TenantUser
from ._models import TenantUserMetadata
from ._models import TenantUserPage
from ._models import WebhookEvent
from ._models import WebhookEventProcessStateChanged
from ._models import WebhookEventProcessStateChangedData
from ._models import WebhookEventTaskStateChanged
from ._models import WebhookEventTaskStateChangedData
from ._models import Worker

from ._enums import AuditedObjectType
from ._enums import AuthenticationType
from ._enums import LogLevel
from ._enums import PagedObjectType
from ._enums import PrincipalType
from ._enums import ProcessElementValueType
from ._enums import ProcessState
from ._enums import RobotAssetArchitecture
from ._enums import RobotAssetPlatform
from ._enums import RobotAssetType
from ._enums import RobotFilterContext
from ._enums import RobotSourceType
from ._enums import TaskElementValueType
from ._enums import TaskState
from ._enums import WebhookType
from ._patch import __all__ as _patch_all
from ._patch import *  # pylint: disable=unused-wildcard-import
from ._patch import patch_sdk as _patch_sdk

__all__ = [
    "AbstractAudited",
    "Authentication",
    "AuthenticationEngineCertificate",
    "AuthenticationEngineCertificateTls",
    "AuthenticationEngineToken",
    "DefaultError",
    "DefaultErrorInfo",
    "JsonFormsValue",
    "Log",
    "Page",
    "PageMetadata",
    "Principal",
    "PrincipalApplication",
    "PrincipalPage",
    "PrincipalUser",
    "Process",
    "ProcessChangeInitiatorCommand",
    "ProcessDefinitionSummary",
    "ProcessDeleteElementCommand",
    "ProcessElementValue",
    "ProcessElementValueNumber",
    "ProcessElementValueString",
    "ProcessPage",
    "ProcessPageItem",
    "ProcessSaveElementCommand",
    "ProcessSaveEntityDataCommand",
    "ProcessSaveEntityDocumentResponseCommand",
    "RelatedProcess",
    "Robot",
    "RobotPage",
    "RobotSourceFile",
    "Task",
    "TaskAssignCommand",
    "TaskDefinitionSummary",
    "TaskDeleteElementCommand",
    "TaskDeleteElementValueDocumentCommand",
    "TaskElementValue",
    "TaskElementValueDocument",
    "TaskElementValueDocumentItem",
    "TaskElementValueNumber",
    "TaskElementValueObject",
    "TaskElementValuePrincipal",
    "TaskElementValuePrincipalItem",
    "TaskElementValueString",
    "TaskPage",
    "TaskPageItem",
    "TaskSaveElementCommand",
    "TaskSaveJsonFormsValueDataCommand",
    "TaskSaveJsonFormsValueDocumentResponseCommand",
    "TenantUser",
    "TenantUserMetadata",
    "TenantUserPage",
    "WebhookEvent",
    "WebhookEventProcessStateChanged",
    "WebhookEventProcessStateChangedData",
    "WebhookEventTaskStateChanged",
    "WebhookEventTaskStateChangedData",
    "Worker",
    "AuditedObjectType",
    "AuthenticationType",
    "LogLevel",
    "PagedObjectType",
    "PrincipalType",
    "ProcessElementValueType",
    "ProcessState",
    "RobotAssetArchitecture",
    "RobotAssetPlatform",
    "RobotAssetType",
    "RobotFilterContext",
    "RobotSourceType",
    "TaskElementValueType",
    "TaskState",
    "WebhookType",
]
__all__.extend([p for p in _patch_all if p not in __all__])
_patch_sdk()



#
# MIT License
#
# Copyright (c) 2022 KuFlow
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import math
import re
from abc import ABC, abstractmethod
from datetime import date, datetime
from typing import Any, Dict, List, Optional, Union

from ..models import (
    JsonFormsFile,
    JsonFormsPrincipal,
    PrincipalType,
    Task,
    TaskPageItem,
    TaskSaveJsonFormsValueDataCommand,
)


JsonFormsSimpleType = Union[str, int, float, bool, date, JsonFormsPrincipal, JsonFormsFile]

ContainerArrayType = List["ComplexType"]

ContainerRecordType = Dict[str, "ComplexType"]

ContainerType = Union[ContainerArrayType, ContainerRecordType]

ComplexType = Union[JsonFormsSimpleType, ContainerType]

JsonFormsModels = Union[Task, TaskPageItem, TaskSaveJsonFormsValueDataCommand]


class JsonFormDataAccessor(ABC):
    @abstractmethod
    def get_data(self) -> Optional[Dict[str, Any]]:
        raise NotImplementedError

    @abstractmethod
    def set_data(self, data: Optional[Dict[str, Any]]):
        raise NotImplementedError


class JsonFormsProperty:
    def __init__(self, container: ContainerType, path: str, value: ComplexType):
        self.container = container
        self.path = path
        self.value = value


def get_json_forms_property_as_str(accessor: JsonFormDataAccessor, property_path: str) -> str:
    """
    Get a json property as "str" following the "propertyPath" passed.

    Arguments:
        accessor: Json form data accessor.
        property_path: Property path to find. ie: "user.name" or "users.0.name"

    Return:
        the property value if exists.

    Raises:
        ValueError: If property value doesn't exist or has incorrect format
    """
    value = find_json_forms_property_as_str(accessor, property_path)
    if value is None:
        raise ValueError("Property value doesn't exist")
    return value


def find_json_forms_property_as_str(accessor: JsonFormDataAccessor, property_path: str) -> Optional[str]:
    """
    Try to find a json property as "str" following the "propertyPath" passed.

    Arguments:
        accessor: Json form data accessor.
        property_path: Property path to find. ie: "user.name" or "users.0.name"

    Return:
        The property value if exists.

    Raises:
        ValueError: If property value has incorrect format
    """
    value = find_json_forms_property_value(accessor, property_path)
    if value is not None:
        return str(value)
    return None


def get_json_forms_property_as_int(accessor: JsonFormDataAccessor, property_path: str) -> int:
    """
    Get a json property as "int" following the "propertyPath" passed.

    Arguments:
        accessor: Json form data accessor.
        property_path: Property path to find. ie: "user.name" or "users.0.name"

    Return:
        the property value if exists.

    Raises:
        ValueError: If property value doesn't exist or has incorrect format
    """
    value = find_json_forms_property_as_int(accessor, property_path)
    if value is None:
        raise ValueError("Property value doesn't exist")
    return value


def find_json_forms_property_as_int(accessor: JsonFormDataAccessor, property_path: str) -> Optional[int]:
    """
    Try to find a json property as "int" following the "propertyPath" passed.

    Arguments:
        accessor: Json form data accessor.
        property_path: Property path to find. ie: "user.name" or "users.0.name"

    Return:
        The property value if exists.

    Raises:
        ValueError: If property value has incorrect format
    """
    value = find_json_forms_property_value(accessor, property_path)
    if value is None:
        return None

    if isinstance(value, int):
        if not math.isnan(value):
            return value

    if isinstance(value, str):
        try:
            return int(value)
        except ValueError:
            pass

    raise ValueError(f"Property {property_path} is not a int")


def get_json_forms_property_as_float(accessor: JsonFormDataAccessor, property_path: str) -> float:
    """
    Get a json property as "float" following the "propertyPath" passed.

    Arguments:
        accessor: Json form data accessor.
        property_path: Property path to find. ie: "user.name" or "users.0.name"

    Return:
        the property value if exists.

    Raises:
        ValueError: If property value doesn't exist or has incorrect format
    """
    value = find_json_forms_property_as_float(accessor, property_path)
    if value is None:
        raise ValueError("Property value doesn't exist")
    return value


def find_json_forms_property_as_float(accessor: JsonFormDataAccessor, property_path: str) -> Optional[float]:
    """
    Try to find a json property as "float" following the "propertyPath" passed.

    Arguments:
        accessor: Json form data accessor.
        property_path: Property path to find. ie: "user.name" or "users.0.name"

    Return:
        The property value if exists.

    Raises:
        ValueError: If property value has incorrect format
    """
    value = find_json_forms_property_value(accessor, property_path)
    if value is None:
        return None

    if isinstance(value, int):
        if not math.isnan(value):
            return float(value)

    if isinstance(value, float):
        if not math.isnan(value):
            return value

    if isinstance(value, str):
        try:
            return float(value)
        except ValueError:
            pass

    raise ValueError(f"Property {property_path} is not a float")


def get_json_forms_property_as_date(accessor: JsonFormDataAccessor, property_path: str) -> date:
    """
    Get a json property as "date" following the "propertyPath" passed.

    Arguments:
        accessor: Json form data accessor.
        property_path: Property path to find. ie: "user.name" or "users.0.name"

    Return:
        the property value if exists.

    Raises:
        ValueError: If property value doesn't exist or has incorrect format
    """
    value = find_json_forms_property_as_date(accessor, property_path)
    if value is None:
        raise ValueError("Property value doesn't exist")
    return value


def find_json_forms_property_as_date(accessor: JsonFormDataAccessor, property_path: str) -> Optional[date]:
    """
    Try to find a json property as "date" following the "propertyPath" passed.

    Arguments:
        accessor: Json form data accessor.
        property_path: Property path to find. ie: "user.name" or "users.0.name"

    Return:
        The property value if exists.

    Raises:
        ValueError: If property value has incorrect format
    """
    value = find_json_forms_property_value(accessor, property_path)
    if value is None:
        return None

    if isinstance(value, date):
        return value

    if isinstance(value, str):
        try:
            return date.fromisoformat(value)
        except ValueError:
            pass

    raise ValueError(f"Property {property_path} is not a date following ISO 8601 format")


def get_json_forms_property_as_datetime(accessor: JsonFormDataAccessor, property_path: str) -> datetime:
    """
    Get a json property as "datetime" following the "propertyPath" passed.

    Arguments:
        accessor: Json form data accessor.
        property_path: Property path to find. ie: "user.name" or "users.0.name"

    Return:
        the property value if exists.

    Raises:
        ValueError: If property value doesn't exist or has incorrect format
    """
    value = find_json_forms_property_as_datetime(accessor, property_path)
    if value is None:
        raise ValueError("Property value doesn't exist")
    return value


def find_json_forms_property_as_datetime(accessor: JsonFormDataAccessor, property_path: str) -> Optional[datetime]:
    """
    Try to find a json property as "datetime" following the "propertyPath" passed.

    Arguments:
        accessor: Json form data accessor.
        property_path: Property path to find. ie: "user.name" or "users.0.name"

    Return:
        The property value if exists.

    Raises:
        ValueError: If property value has incorrect format
    """
    value = find_json_forms_property_value(accessor, property_path)
    if value is None:
        return None

    if isinstance(value, datetime):
        return value

    if isinstance(value, str):
        try:
            return datetime.fromisoformat(value)
        except ValueError:
            pass

    raise ValueError(f"Property {property_path} is not a date-time following ISO 8601 format")


def get_json_forms_property_as_json_forms_file(accessor: JsonFormDataAccessor, property_path: str) -> JsonFormsFile:
    """
    Get a json property as "JsonFormsFile" following the "propertyPath" passed.

    Arguments:
        accessor: Json form data accessor.
        property_path: Property path to find. ie: "user.name" or "users.0.name"

    Return:
        the property value if exists.

    Raises:
        ValueError: If property value doesn't exist or has incorrect format
    """
    value = find_json_forms_property_as_json_forms_file(accessor, property_path)
    if value is None:
        raise ValueError("Property value doesn't exist")
    return value


def find_json_forms_property_as_json_forms_file(
    accessor: JsonFormDataAccessor, property_path: str
) -> Optional[JsonFormsFile]:
    """
    Try to find a json property as "JsonFormsFile" following the "propertyPath" passed.

    Arguments:
        accessor: Json form data accessor.
        property_path: Property path to find. ie: "user.name" or "users.0.name"

    Return:
        The property value if exists.

    Raises:
        ValueError: If property value has incorrect format
    """
    value = find_json_forms_property_value(accessor, property_path)
    if value is None:
        return None

    json_forms_file = try_parse_json_forms_file(value)
    if json_forms_file is None:
        raise ValueError(f"Property {property_path} is not a file")

    return json_forms_file


def get_json_forms_property_as_json_forms_principal(
    accessor: JsonFormDataAccessor, property_path: str
) -> JsonFormsPrincipal:
    """
    Get a json property as "JsonFormsPrincipal" following the "propertyPath" passed.

    Arguments:
        accessor: Json form data accessor.
        property_path: Property path to find. ie: "user.name" or "users.0.name"

    Return:
        the property value if exists.

    Raises:
        ValueError: If property value doesn't exist or has incorrect format
    """
    value = find_json_forms_property_as_json_forms_principal(accessor, property_path)
    if value is None:
        raise ValueError("Property value doesn't exist")
    return value


def find_json_forms_property_as_json_forms_principal(
    accessor: JsonFormDataAccessor, property_path: str
) -> Optional[JsonFormsPrincipal]:
    """
    Try to find a json property as "JsonFormsPrincipal" following the "propertyPath" passed.

    Arguments:
        accessor: Json form data accessor.
        property_path: Property path to find. ie: "user.name" or "users.0.name"

    Return:
        The property value if exists.

    Raises:
        ValueError: If property value has incorrect format
    """
    value = find_json_forms_property_value(accessor, property_path)
    if value is None:
        return None

    json_forms_file = try_parse_json_forms_principal(value)
    if json_forms_file is None:
        raise ValueError(f"Property {property_path} is not a principal")

    return json_forms_file


def get_json_forms_property_as_list(accessor: JsonFormDataAccessor, property_path: str) -> list:
    """
    Get a json property as "list" following the "propertyPath" passed.

    Arguments:
        accessor: Json form data accessor.
        property_path: Property path to find. ie: "user.name" or "users.0.name"

    Return:
        the property value if exists.

    Raises:
        ValueError: If property value doesn't exist or has incorrect format
    """
    value = find_json_forms_property_as_list(accessor, property_path)
    if value is None:
        raise ValueError("Property value doesn't exist")
    return value


def find_json_forms_property_as_list(accessor: JsonFormDataAccessor, property_path: str) -> Optional[list]:
    """
    Try to find a json property as "list" following the "propertyPath" passed.

    Arguments:
        accessor: Json form data accessor.
        property_path: Property path to find. ie: "user.name" or "users.0.name"

    Return:
        The property value if exists.

    Raises:
        ValueError: If property value has incorrect format
    """
    value = find_json_forms_property_value(accessor, property_path)
    if value is None:
        return None

    if isinstance(value, list):
        return value

    raise ValueError(f"Property {property_path} is not a list")


def get_json_forms_property_as_dict(accessor: JsonFormDataAccessor, property_path: str) -> dict:
    """
    Get a json property as "dict" following the "propertyPath" passed.

    Arguments:
        accessor: Json form data accessor.
        property_path: Property path to find. ie: "user.name" or "users.0.name"

    Return:
        the property value if exists.

    Raises:
        ValueError: If property value doesn't exist or has incorrect format
    """
    value = find_json_forms_property_as_dict(accessor, property_path)
    if value is None:
        raise ValueError("Property value doesn't exist")
    return value


def find_json_forms_property_as_dict(accessor: JsonFormDataAccessor, property_path: str) -> Optional[dict]:
    """
    Try to find a json property as "dict" following the "propertyPath" passed.

    Arguments:
        accessor: Json form data accessor.
        property_path: Property path to find. ie: "user.name" or "users.0.name"

    Return:
        The property value if exists.

    Raises:
        ValueError: If property value has incorrect format
    """
    value = find_json_forms_property_value(accessor, property_path)
    if value is None:
        return None

    if isinstance(value, dict):
        return value

    raise ValueError(f"Property {property_path} is not a dict")


def find_json_forms_property_value(
    accessor: JsonFormDataAccessor,
    property_path: str,
    create_missing_parents: bool = True,
) -> Optional[ComplexType]:
    property = find_json_forms_property(accessor, property_path, create_missing_parents)
    if property is not None:
        return property.value
    return None


def update_json_forms_property(
    accessor: JsonFormDataAccessor,
    property_path: str,
    value: Optional[JsonFormsSimpleType],
) -> None:
    """
    Update a json forms data property in the task passed following the "property_path".

    Arguments:
        accessor: Json form data accessor.
        property_path: Property path to find. ie: "user.name" or "users.0.name"
        value: Value to update

    Raises:
        ValueError: If property value has incorrect format
    """
    property = find_json_forms_property(accessor, property_path, True)

    value = transform_json_forms_property_value(value)

    if property is None:
        raise ValueError(f"Property {property_path} doesn't exist")

    json_forms_property_container = property.container
    json_forms_property_path = property.path
    if isinstance(json_forms_property_container, list):
        if not json_forms_property_path.isdigit():
            raise ValueError(f"Incorrect property path {json_forms_property_path}, parent path is not a List")

        json_forms_property_path_index = int(json_forms_property_path)
        if value is not None:
            json_forms_property_container[json_forms_property_path_index] = value
        else:
            json_forms_property_container.pop(json_forms_property_path_index)
    elif isinstance(json_forms_property_container, dict):
        if value is not None:
            json_forms_property_container[json_forms_property_path] = value
        else:
            del json_forms_property_container[json_forms_property_path]
    else:
        raise ValueError(f"Incorrect property path {json_forms_property_path}")


# flake8: noqa: C901
def find_json_forms_property(
    accessor: JsonFormDataAccessor,
    property_path: str,
    create_missing_parents: bool = False,
) -> Optional[JsonFormsProperty]:
    """
    Try to find a json property following the "property_path" passed.

    Arguments:
        accessor: Json form data accessor.
        property_path: Property path to find. ie: "user.name" or "users.0.name"
        create_missing_parents: Create parent paths if are missing

    Return:
        The property value if exists.

    Raises:
        ValueError: If property value has incorrect format or a wrong path
    """
    property_data = get_json_forms_value_data(accessor, create_missing_parents)
    if property_data is None:
        return None

    property_container: ContainerType = property_data
    property_value_path = ""
    property_value: Optional[ComplexType] = None

    paths = property_path.split(".")
    for idx, path in enumerate(paths):
        property_value_path = path
        if property_value_path == "":
            continue

        property_value_path_as_integer = -1

        if is_json_forms_type_container_array(property_container):
            if not property_value_path.isdigit():
                raise ValueError(f"Wrong list index {property_value_path}")

            property_value_path_as_integer = int(property_value_path)
            if property_value_path_as_integer < 0 or property_value_path_as_integer > len(property_container):
                return None
            elif property_value_path_as_integer == len(property_container):
                if not create_missing_parents:
                    return None

                property_value = None
            else:
                property_value = property_container[property_value_path_as_integer]
        elif is_json_forms_type_container_record(property_container):
            if property_value_path not in property_container:
                if not create_missing_parents:
                    return None

            property_value = property_container.get(property_value_path)
        else:
            return None

        if property_value is None:
            if not create_missing_parents:
                return None

            if idx + 1 < len(paths):
                path_next = paths[idx + 1]
                if path_next.isdigit():
                    property_value = []
                else:
                    property_value = {}

                if is_json_forms_type_container_array(property_container):
                    if property_value_path_as_integer != len(property_container):
                        raise ValueError(f"Wrong list index {property_value_path}")

                    property_container.append(property_value)
                elif is_json_forms_type_container_record(property_container):
                    property_container[property_value_path] = property_value

        if property_value is not None and is_json_forms_type_container(property_value) and idx + 1 < len(paths):
            property_container = property_value
            property_value = None

    return JsonFormsProperty(container=property_container, path=property_value_path, value=property_value)


def transform_json_forms_property_value(
    value: Optional[JsonFormsSimpleType],
) -> Optional[JsonFormsSimpleType]:
    if value is None:
        return None
    elif isinstance(value, JsonFormsPrincipal):
        return generate_value_for_json_forms_principal(value)
    elif isinstance(value, JsonFormsFile):
        return generate_value_for_json_forms_file(value)
    elif isinstance(value, date):
        return value.isoformat()
    elif isinstance(value, datetime):
        return value.isoformat()
    elif isinstance(value, (int, float, bool, str)):
        return value

    raise Exception(f"Unsupported value {value}")


def generate_value_for_json_forms_principal(principal: JsonFormsPrincipal) -> str:
    return f"kuflow-principal:id={principal.id};type={principal.type};name={principal.name};"


def generate_value_for_json_forms_file(file: JsonFormsFile) -> str:
    return f"kuflow-file:uri={file.uri};type={file.type};size={file.size};name={file.name};"


def try_parse_json_forms_file(value: ComplexType) -> Optional[JsonFormsFile]:
    if value is None:
        return None

    if not isinstance(value, str):
        return None

    if not value.startswith("kuflow-file:"):
        return None

    value_transformed = value.replace("kuflow-file:", "")

    matches = re.findall(".*?=.*?;", value_transformed)
    if matches is None or len(matches) != 4:
        return None

    uri: Optional[str] = None
    type: Optional[str] = None
    name: Optional[str] = None
    size: Optional[int] = None

    for match in matches:
        match = match[:-1]
        key, value = match.split("=")
        if key == "uri":
            uri = str(value)
        elif key == "type":
            type = str(value)
        elif key == "name":
            name = str(value)
        elif key == "size":
            size = int(value)

    return JsonFormsFile(
        uri=uri,
        type=type,
        name=name,
        size=size,
    )


def try_parse_json_forms_principal(value: ComplexType) -> Optional[JsonFormsPrincipal]:
    if value is None:
        return None

    if not isinstance(value, str):
        return None

    if not value.startswith("kuflow-principal:"):
        return None

    value_transformed = value.replace("kuflow-principal:", "")

    matches = re.findall(".*?=.*?;", value_transformed)
    if matches is None or len(matches) != 3:
        return None

    id: Optional[str] = None
    type: Optional[PrincipalType] = None
    name: Optional[str] = None

    for match in matches:
        match = match[:-1]
        key, value = match.split("=")
        if key == "id":
            id = str(value)
        elif key == "type":
            type = PrincipalType(value)
        elif key == "name":
            name = str(value)

    return JsonFormsPrincipal(
        id=id,
        type=type,
        name=name,
    )


def is_json_forms_type_container(value: Optional[ComplexType]) -> bool:
    return is_json_forms_type_container_record(value) or is_json_forms_type_container_array(value)


def is_json_forms_type_container_record(value: Optional[ComplexType]) -> bool:
    return isinstance(value, dict)


def is_json_forms_type_container_array(value: Optional[ComplexType]) -> bool:
    return isinstance(value, list)


def get_json_forms_value_data(accessor: JsonFormDataAccessor, create_missing_parents: bool) -> Optional[ContainerType]:
    data: Optional[ContainerType] = accessor.get_data()

    if data is None and create_missing_parents:
        data = {}
        accessor.set_data(data)

    return data
